#define _BSD_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <modbus/modbus.h>

typedef struct {
	modbus_param_t	mb_param;
	int		slave_id;

	GtkWidget       *window;

	GtkWidget	*address_entry;
	GtkWidget	*set_voltage_entry;
	GtkWidget	*set_current_entry;
	GtkWidget	*voltage_entry;
	GtkWidget	*current_entry;
	GtkWidget	*flags_entry;
	GtkWidget	*temp_entry;
	GtkWidget	*device_entry;
	GtkWidget	*voltage_calibration_entry;
	GtkWidget	*current_calibration_entry;
	GtkWidget	*turnon_button;
	GtkWidget	*turnoff_button;
	GtkWidget	*reset_button;

	GPtrArray	*active_buttons;
	GtkWidget	*connect;
	GtkWidget	*request_info;
} RectControlApp;

RectControlApp rectcontrolapp;
RectControlApp *rectcontrol = &rectcontrolapp;
guint request_info_worker_id;
GKeyFile *conf;

struct current_state {
	double	voltage;
	double	current;
	uint8_t	flags;
	double	temppfc;
	double	tempdc;
};

double code_to_temp(uint16_t code)
{
	int i;
	const double temp_table[][2]={ {350, -48.7},
				       {460, 0.0},
				       {517, 27.0},
				       {575, 58.6},
				       {619, 87.0},
				       {638, 100.0},
				       {668, 124.6},
				       {700, 155.9}};
	if (code<temp_table[0][0])
		return -99;
	for (i=0; i<sizeof(temp_table)/sizeof(temp_table[0][0])/2 - 1; i++)
		if (code >= temp_table[i][0] && code <= temp_table[i+1][0])
			return temp_table[i][1] +
				(code-temp_table[i][0]) *
				(temp_table[i+1][1]-temp_table[i][1]) /
				(temp_table[i+1][0]-temp_table[i][0]);
	return 999;
}

int get_current_state( struct current_state * cs)
{
	uint16_t buf[8];
	
	if (read_holding_registers( &rectcontrol->mb_param,
				    rectcontrol->slave_id, 1000, 8, buf ) < 0)
		return 0;
	
	fprintf(stderr, "%d %d 0x%x %d %d 0x%x %d %d\n", buf[0], buf[1], buf[2],
		buf[3], buf[4], buf[5], buf[6], buf[7] );
	
	cs->voltage = buf[3]/10.;
	cs->current = buf[4]/10.;
	cs->flags = buf[5];
	cs->temppfc = code_to_temp(buf[6]);
	cs->tempdc = code_to_temp(buf[7]);
	return 1;
}

void set_button_sensitivity(gpointer data, gpointer user_data)
{
	gtk_widget_set_sensitive(GTK_WIDGET( data ), (gboolean)user_data);
}

void 
on_window_destroy (GtkObject *object, gpointer user_data)
{
	g_key_file_set_string(conf, "General", "Device",
			      gtk_entry_get_text(GTK_ENTRY(rectcontrol->device_entry)));
	g_key_file_set_string(conf, "General", "Address",
			      gtk_entry_get_text(GTK_ENTRY(rectcontrol->address_entry)));
	gtk_main_quit ();
}

void 
on_connect_clicked (GtkObject *object, gpointer user_data)
{
    	modbus_init_rtu( &rectcontrol->mb_param,
			 gtk_entry_get_text(GTK_ENTRY( rectcontrol->device_entry )), 9600, "none", 8, 1 );
#if 0
	modbus_set_debug( &rectcontrol->mb_param, 1 );
#endif
	if (modbus_connect( &rectcontrol->mb_param ) != -1) {
		gtk_widget_set_sensitive(GTK_WIDGET( rectcontrol->connect ), FALSE);
		g_ptr_array_foreach( rectcontrol->active_buttons, set_button_sensitivity, (gpointer)TRUE );
	}
}

void 
on_disconnect_clicked (GtkObject *object, gpointer user_data)
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(rectcontrol->request_info), FALSE);
	modbus_close( &rectcontrol->mb_param );
	gtk_widget_set_sensitive(GTK_WIDGET( rectcontrol->connect ), TRUE);
	g_ptr_array_foreach( rectcontrol->active_buttons, set_button_sensitivity, FALSE );
}

void 
on_set_voltage_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 1000,
				(atof(gtk_entry_get_text(
					     GTK_ENTRY( rectcontrol->set_voltage_entry ))) * 100 + 5 )/10);
}

void 
on_set_current_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 1001,
				(atof(gtk_entry_get_text(
					      GTK_ENTRY( rectcontrol->set_current_entry ))) * 100 + 5 )/10);
}

void 
on_set_control_clicked (GtkObject *object, gpointer user_data)
{
	int code;
	GtkWidget *o = GTK_WIDGET( object );
	if (o==rectcontrol->turnon_button)
		code = 0;
	else if (o==rectcontrol->turnoff_button)
		code = 1;
	else if (o==rectcontrol->reset_button)
		code = 2;
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 1002,
				code );
}

gboolean request_info_worker(gpointer data)
{
	char		str[40];
	char		na[]="n/a";
	int		result = 1;
	struct current_state cs;
	result = get_current_state( &cs );

	snprintf(str, sizeof(str), "%.1f", cs.voltage);
	gtk_entry_set_text(GTK_ENTRY( rectcontrol->voltage_entry ), result?str:na);
	snprintf(str, sizeof(str), "%.1f", cs.current);
	gtk_entry_set_text(GTK_ENTRY( rectcontrol->current_entry ), result?str:na);
	switch ((cs.flags>>4) & 0x03) {
	case 0:
		strcpy(str, "LEDTEST");
		break;
	case 1:
		strcpy(str, "OFF");
		break;
	case 2:
		strcpy(str, "ON");
		break;
	case 3:
		strcpy(str, "PROTECT");
		break;
	}
	if (cs.flags>>6 & 1)
		strcat(str, " OVERCURR");
	if (cs.flags>>3 & 1)
		strcat(str, " ACERR");
	if (cs.flags>>2 & 1)
		strcat(str, " VOUTOVER");
	if (cs.flags>>1 & 1)
		strcat(str, " VOUTERR");
	if (cs.flags>>0 & 1)
		strcat(str, " TEMP");
	gtk_entry_set_text(GTK_ENTRY( rectcontrol->flags_entry ), result?str:na);
	snprintf(str, sizeof(str), "PFC:%3.0f   DC:%3.0f", cs.temppfc, cs.tempdc);
	gtk_entry_set_text(GTK_ENTRY( rectcontrol->temp_entry ), result?str:na);
	return TRUE;
}

void 
on_request_info_toggled (GtkObject *object, gpointer user_data)
{
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(object)) == TRUE)
		request_info_worker_id = g_idle_add(request_info_worker, NULL);
	else
		g_source_remove( request_info_worker_id );
}

void 
on_address_entry_changed (GtkObject *object, gpointer user_data)
{
	rectcontrol->slave_id = atoi( gtk_entry_get_text(GTK_ENTRY( object )) );
}

void
on_ac_min_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 900, 0 );
}

void
on_ac_preduce_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 900, 1 );
}

void
on_ac_max_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 900, 2 );
}

void
on_vol_cal_start_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 1000,
				480 );
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 901, 128 );
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 902, 32767 );
}

void
on_cur_cal_start_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 1001,
				180 );
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 903, 32767 );
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 904, 128 );
}

void
on_vol_cal_clicked (GtkObject *object, gpointer user_data)
{
	double current_voltage;
	uint8_t set_cc;
	uint16_t meas_cc;
	struct current_state cs;
	current_voltage = atof(gtk_entry_get_text(
				       GTK_ENTRY( rectcontrol->voltage_calibration_entry )));
	get_current_state( &cs );
	set_cc = (48/current_voltage - 1)*1000 + 128 + .5;
	meas_cc = (current_voltage/cs.voltage - 1)*1000 + 32767 + .5;
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 901,
				set_cc );
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 902,
				meas_cc );
}

void
on_cur_cal_clicked (GtkObject *object, gpointer user_data)
{
	double current_current;
	uint16_t set_cc;
	uint8_t meas_cc;
	struct current_state cs;
	current_current = atof(gtk_entry_get_text(
				       GTK_ENTRY( rectcontrol->current_calibration_entry )));
	get_current_state( &cs );
	set_cc = (18/current_current - 1)*1000 + 32767 + .5;
	meas_cc = (current_current/cs.current - 1)*1000 + 128 + .5;
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 903,
				set_cc );
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 904,
				meas_cc );
	preset_single_register( &rectcontrol->mb_param,
				rectcontrol->slave_id, 1001,
				250 );
}

int
main (int argc, char *argv[])
{
	GtkBuilder      *builder;
	gchar *conf_path, *conf_path_new, *conf_data;
	const gchar *tmp;
	int newconffile;

	setlocale(LC_ALL, "");
	
	gtk_init (&argc, &argv);
 
	builder = gtk_builder_new ();
	gtk_builder_add_from_file (builder, "rect_control.xml", NULL);
	rectcontrol->window = GTK_WIDGET (gtk_builder_get_object (
						  builder, "mainwindow"));
	rectcontrol->address_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "address_entry"));
	rectcontrol->set_voltage_entry = GTK_WIDGET (gtk_builder_get_object (
							     builder, "set_voltage_entry"));
	rectcontrol->set_current_entry = GTK_WIDGET (gtk_builder_get_object (
							     builder, "set_current_entry"));
	rectcontrol->voltage_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "voltage_entry"));
	rectcontrol->current_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "current_entry"));
	rectcontrol->flags_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "flags_entry"));
	rectcontrol->temp_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "temp_entry"));
	rectcontrol->device_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "device_entry"));
	rectcontrol->voltage_calibration_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "voltage_calibration_entry"));
	rectcontrol->current_calibration_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "current_calibration_entry"));

	rectcontrol->connect = GTK_WIDGET ( gtk_builder_get_object (
						    builder, "connect"));
	rectcontrol->active_buttons = g_ptr_array_new();
	g_ptr_array_add( rectcontrol->active_buttons, gtk_builder_get_object(
				   builder, "disconnect" ));
	g_ptr_array_add( rectcontrol->active_buttons, gtk_builder_get_object(
				   builder, "request_info" ));
	g_ptr_array_add( rectcontrol->active_buttons, gtk_builder_get_object(
				   builder, "set_voltage" ));
	g_ptr_array_add( rectcontrol->active_buttons, gtk_builder_get_object(
				 builder, "set_current" ));
	rectcontrol->turnon_button = GTK_WIDGET( gtk_builder_get_object(
							  builder, "turnon" ));
	g_ptr_array_add( rectcontrol->active_buttons, rectcontrol->turnon_button );
	rectcontrol->turnoff_button = GTK_WIDGET( gtk_builder_get_object(
							  builder, "turnoff" ));
	g_ptr_array_add( rectcontrol->active_buttons, rectcontrol->turnoff_button );
	rectcontrol->reset_button = GTK_WIDGET( gtk_builder_get_object(
							builder, "reset" ));
	g_ptr_array_add( rectcontrol->active_buttons, rectcontrol->reset_button );

	rectcontrol->request_info = GTK_WIDGET (gtk_builder_get_object(builder, "request_info" ));
		
	gtk_builder_connect_signals (builder, NULL);
 
	g_object_unref (G_OBJECT (builder));

	conf = g_key_file_new();
	tmp = g_get_home_dir();
	conf_path = g_build_filename(tmp, ".rect_control", NULL);
	conf_path_new = g_build_filename(tmp, ".rect_control.new", NULL);
	g_free(tmp);
	if (g_key_file_load_from_file(conf, conf_path,
				  G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS,
				      NULL)==TRUE) {
		tmp = g_key_file_get_string(conf, "General", "Device", NULL);
		gtk_entry_set_text(GTK_ENTRY( rectcontrol->device_entry ), tmp);
		g_free(tmp);
		tmp = g_key_file_get_string(conf, "General", "Address", NULL);
		gtk_entry_set_text(GTK_ENTRY( rectcontrol->address_entry ), tmp);
		g_free(tmp);
	}
	
	gtk_widget_show (rectcontrol->window);                
	gtk_main ();

	conf_data = g_key_file_to_data(conf, NULL, NULL);
	newconffile = g_open(conf_path_new, O_WRONLY|O_TRUNC|O_CREAT, 0600);
	if (newconffile==-1)
		perror("Warning: can't save configuration data");
	else {
		write(newconffile, conf_data, strlen(conf_data));
#ifndef NATIVE_WIN32
		if (fsync(newconffile)==-1) {
			close(newconffile);
			perror("Warning: can't sync configuration data");
		} else {
			close(newconffile);
			g_rename(conf_path_new, conf_path);
		}
#else
		close(newconffile);
		g_rename(conf_path_new, conf_path);
#endif
	}
	g_free(conf_data);
	g_free(conf_path_new);
	g_free(conf_path);
	g_key_file_free(conf);
		
	return 0;
}

